import requests
from rich.console import Console
console = Console()


class Mixin:
    def ql_get_blockhash(self, block_height):
        query = "query {"
        query += f'search (query: "{block_height}")'
        query += """
                  {
                    blocks (last: 1) {
                    nodes {
                        blockHeight
                        blockHash
                    }
                    }
                }
                
                }
        """
        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                if len (r.json()['data']['search']['blocks']['nodes']) > 0:
                    return r.json()['data']['search']['blocks']['nodes'][0]['blockHash']
                else:
                    return None
            else:
                console.log(r.text)
        except Exception as e:
            console.log(query, e)
            return None


import os
import ast
from dotenv import load_dotenv

load_dotenv()


BRANCH = os.environ.get("BRANCH", "dev")
ENVIRONMENT = os.environ.get("ENVIRONMENT", "prod")
NOTIFIER_API_TOKEN = os.environ.get("NOTIFIER_API_TOKEN")
API_TOKEN = os.environ.get("API_TOKEN")
FASTMAIL_TOKEN = os.environ.get("FASTMAIL_TOKEN")
FALLBACK_URI = os.environ.get("FALLBACK_URI")
MONGO_URI = os.environ.get("MONGO_URI")
ADMIN_CHAT_ID = os.environ.get("ADMIN_CHAT_ID")
MAILTO_LINK = os.environ.get("MAILTO_LINK")
MAILTO_USER = os.environ.get("MAILTO_USER")
GRPC_MAINNET = ast.literal_eval(os.environ["GRPC_MAINNET"])
GRPC_TESTNET = ast.literal_eval(os.environ["GRPC_TESTNET"])



# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient
import os
import sys

sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *
from google.protobuf.json_format import MessageToDict


class Mixin(_SharedConverters):
    def convertNewRelease(self, message) -> list:
        schedule = []
        for entry in message:
            entry_dict = {}
            for descriptor in entry.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, entry)
                if type(value) in self.simple_types:
                    converted_value = self.convertType(value)
                    if converted_value:
                        entry_dict[key] = converted_value
            schedule.append(CCD_NewRelease(**entry_dict))

        return schedule

    def convertBakerKeysEvent(self, message) -> CCD_BakerKeysEvent:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerKeysEvent(**result)

    def convertEffectBakerAdded(self, message) -> CCD_BakerAdded:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            if type(value) == BakerKeysEvent:
                result[key] = self.convertBakerKeysEvent(value)

        return CCD_BakerAdded(**result)

    def convertBakerStakeUpdatedData(self, message) -> CCD_BakerStakeUpdatedData:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerStakeUpdatedData(**result)

    def convertEffectBakerStakeUpdated(self, message) -> CCD_BakerStakeUpdated:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            if type(value) == BakerStakeUpdatedData:
                result[key] = self.convertBakerStakeUpdatedData(value)

        return CCD_BakerStakeUpdated(**result)

    def convertBakerStakeIncreased(self, message) -> CCD_BakerStakeIncreased:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerStakeIncreased(**result)

    def convertBakerStakeDecreased(self, message) -> CCD_BakerStakeDecreased:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerStakeDecreased(**result)

    def convertDelegationStakeIncreased(self, message) -> CCD_DelegationStakeIncreased:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_DelegationStakeIncreased(**result)

    def convertDelegationStakeDecreased(self, message) -> CCD_DelegationStakeDecreased:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_DelegationStakeDecreased(**result)

    def convertBakerSetOpenStatus(self, message) -> CCD_BakerSetOpenStatus:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerSetOpenStatus(**result)

    def convertBakerSetMetadataUrl(self, message) -> CCD_BakerSetMetadataUrl:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerSetMetadataUrl(**result)

    def convertDelegationSetSetRestakeEarnings(
        self, message
    ) -> CCD_DelegationSetRestakeEarnings:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_DelegationSetRestakeEarnings(**result)

    def convertBakerRestakeEarningsUpdated(
        self, message
    ) -> CCD_BakerRestakeEarningsUpdated:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerRestakeEarningsUpdated(**result)

    def convertBakerSetBakingRewardCommission(
        self, message
    ) -> CCD_BakerSetBakingRewardCommission:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            if key == "baking_reward_commission":
                result[key] = self.convertType(value)
        return CCD_BakerSetBakingRewardCommission(**result)

    def convertBakerSetFinalizationRewardCommission(
        self, message
    ) -> CCD_BakerSetFinalizationRewardCommission:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            if key == "finalization_reward_commission":
                result[key] = self.convertType(value)
        return CCD_BakerSetFinalizationRewardCommission(**result)

    def convertBakerSetTransactionFeeCommission(
        self, message
    ) -> CCD_BakerSetTransactionFeeCommission:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            if key == "transaction_fee_commission":
                result[key] = self.convertType(value)
        return CCD_BakerSetTransactionFeeCommission(**result)

    def convertBakerBakerAdded(self, message) -> CCD_BakerAdded:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) == BakerKeysEvent:
                result[key] = self.convertBakerKeysEvent(value)

            elif type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_BakerAdded(**result)

    def convertDelegationSetDelegationTarget(
        self, message
    ) -> CCD_DelegationSetDelegationTarget:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) == DelegationTarget:
                result[key] = self.convertDelegationTarget(value)

        return CCD_DelegationSetDelegationTarget(**result)

    def convertBakerConfiguredEvents(self, message) -> list:
        events = []
        for entry in message:
            for descriptor in entry.DESCRIPTOR.fields:
                result = {}
                key, value = self.get_key_value_from_descriptor(descriptor, entry)

                if self.valueIsEmpty(value, key, message):
                    pass

                else:
                    if type(value) == BakerEvent.BakerStakeIncreased:
                        result[key] = self.convertBakerStakeIncreased(value)
                        events.append(result)

                    elif type(value) == BakerEvent.BakerStakeDecreased:
                        result[key] = self.convertBakerStakeDecreased(value)
                        events.append(result)

                    elif type(value) == BakerEvent.BakerSetMetadataUrl:
                        result[key] = self.convertBakerSetMetadataUrl(value)
                        events.append(result)

                    elif type(value) == BakerEvent.BakerSetOpenStatus:
                        result[key] = self.convertBakerSetOpenStatus(value)
                        events.append(result)

                    elif type(value) == BakerEvent.BakerRestakeEarningsUpdated:
                        result[key] = self.convertBakerRestakeEarningsUpdated(value)
                        events.append(result)

                    elif type(value) == BakerEvent.BakerSetBakingRewardCommission:
                        result[key] = self.convertBakerSetBakingRewardCommission(value)
                        events.append(result)

                    elif type(value) == BakerEvent.BakerSetTransactionFeeCommission:
                        result[key] = self.convertBakerSetTransactionFeeCommission(
                            value
                        )
                        events.append(result)

                    elif type(value) == BakerEvent.BakerSetFinalizationRewardCommission:
                        result[key] = self.convertBakerSetFinalizationRewardCommission(
                            value
                        )
                        events.append(result)

                    elif type(value) == BakerKeysEvent:
                        result[key] = self.convertBakerKeysEvent(value)
                        events.append(result)

                    elif key == "baker_removed":
                        result[key] = self.convertType(value)
                        events.append(result)

                    elif key == "baker_added":
                        result[key] = self.convertBakerBakerAdded(value)
                        events.append(result)

        return events

    def convertDelegationConfiguredEvents(self, message) -> list:
        events = []
        for entry in message:
            for descriptor in entry.DESCRIPTOR.fields:
                result = {}
                key, value = self.get_key_value_from_descriptor(descriptor, entry)
                if MessageToDict(value) == {}:
                    pass
                else:
                    if type(value) == DelegationEvent.DelegationStakeIncreased:
                        result[key] = self.convertDelegationStakeIncreased(value)
                        events.append(result)

                    elif type(value) == DelegationEvent.DelegationStakeDecreased:
                        result[key] = self.convertDelegationStakeDecreased(value)
                        events.append(result)

                    elif type(value) == DelegationEvent.DelegationSetDelegationTarget:
                        result[key] = self.convertDelegationSetDelegationTarget(value)
                        events.append(result)

                    elif type(value) == DelegationEvent.DelegationSetRestakeEarnings:
                        result[key] = self.convertDelegationSetSetRestakeEarnings(value)
                        events.append(result)

                    elif key == "delegation_removed":
                        result[key] = self.convertType(value)
                        events.append(result)

                    elif key == "delegation_removed":
                        result[key] = self.convertType(value)
                        events.append(result)

        return events

    def convertEffectBakerConfigured(self, message) -> CCD_BakerConfigured:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if key == "events":
                result[key] = self.convertBakerConfiguredEvents(value)

        return CCD_BakerConfigured(**result)

    def convertEffectDelegationConfigured(self, message) -> CCD_DelegationConfigured:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if key == "events":
                result[key] = self.convertDelegationConfiguredEvents(value)

        return CCD_DelegationConfigured(**result)

    def convertEffectAccountTransfer(self, message) -> CCD_AccountTransfer:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if MessageToDict(value) == {}:
                pass
            else:
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)

        return CCD_AccountTransfer(**result)

    def convertEffectAccountTransferWithSchedule(
        self, message
    ) -> CCD_TransferredWithSchedule:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if key == "amount":
                result[key] = self.convertNewRelease(value)

            elif type(value) in self.simple_types:
                converted_value = self.convertType(value)
                if converted_value:
                    result[key] = converted_value

        return CCD_TransferredWithSchedule(**result)

    def convertEffectContractInitializedEvent(
        self, message
    ) -> CCD_ContractInitializedEvent:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if key == "events":
                result[key] = self.convertEvents(value)

            elif type(value) in self.simple_types:
                result[key] = self.convertType(value)
                # if converted_value:
                #      = converted_value

        return CCD_ContractInitializedEvent(**result)

    def convertEffectContractUpdateIssued(self, message) -> CCD_ContractUpdateIssued:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if key == "effects":
                result[key] = self.convertUpdateEvents(value)

        return CCD_ContractUpdateIssued(**result)

    def convertCredentialsUpdated(
        self, message
    ) -> CCD_AccountTransactionEffects_CredentialsUpdated:
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if key in ["new_cred_ids", "removed_cred_ids"]:
                result[key] = self.convertCredentialRegistrationIdEntries(value)

            elif type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_AccountTransactionEffects_CredentialsUpdated(**result)

    def convertEffectAccountEncryptedAmountTransferred(
        self, message
    ) -> CCD_AccountTransactionEffects_EncryptedAmountTransferred:
        result = {}
        _type = None
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) in [EncryptedAmountRemovedEvent, NewEncryptedAmountEvent]:
                result[key] = self.convertTypeWithSingleValues(value)

        return CCD_AccountTransactionEffects_EncryptedAmountTransferred(**result)

    def convertEffectAccountTransferredToPublic(
        self, message
    ) -> CCD_AccountTransactionEffects_TransferredToPublic:
        result = {}
        _type = None
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) in [EncryptedAmountRemovedEvent]:
                result[key] = self.convertTypeWithSingleValues(value)

        return CCD_AccountTransactionEffects_TransferredToPublic(**result)

    def convertAccountTransactionEffects(
        self, message
    ) -> CCD_AccountTransactionEffects:
        result = {}
        _type = {"type": "account_transaction"}
        for field, value in message.ListFields():
            key = field.name
            _outcome = "success"
            try:
                if value.HasField("reject_reason"):
                    _outcome = "reject"
                    result[key], type_contents = self.convertRejectReasonNone(value)
                    _type.update({"contents": type_contents})
            except:  # noqa F403
                if _outcome == "success":
                    if type(value) == ContractInitializedEvent:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectContractInitializedEvent(value)

                    elif type(value) == AccountTransactionEffects.ContractUpdateIssued:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectContractUpdateIssued(value)

                    elif type(value) == AccountTransactionEffects.AccountTransfer:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectAccountTransfer(value)

                    elif type(value) == BakerEvent.BakerAdded:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectBakerAdded(value)

                    elif type(value) == BakerId:
                        _type.update({"contents": key})
                        result[key] = self.convertType(value)

                    elif type(value) == AccountTransactionEffects.BakerStakeUpdated:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectBakerStakeUpdated(value)

                    elif type(value) == BakerEvent.BakerRestakeEarningsUpdated:
                        _type.update({"contents": key})
                        result[key] = self.convertTypeWithSingleValues(value)

                    elif type(value) == BakerKeysEvent:
                        _type.update({"contents": key})
                        result[key] = self.convertBakerKeysEvent(value)

                    elif (
                        type(value)
                        == AccountTransactionEffects.EncryptedAmountTransferred
                    ):
                        _type.update({"contents": key})
                        result[key] = (
                            self.convertEffectAccountEncryptedAmountTransferred(value)
                        )

                    elif type(value) == EncryptedSelfAmountAddedEvent:
                        _type.update({"contents": key})
                        result[key] = CCD_EncryptedSelfAmountAddedEvent(
                            **self.convertTypeWithSingleValues(value)
                        )

                    elif type(value) == AccountTransactionEffects.TransferredToPublic:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectAccountTransferredToPublic(
                            value
                        )

                    elif (
                        type(value) == AccountTransactionEffects.TransferredWithSchedule
                    ):
                        _type.update({"contents": key})
                        result[key] = self.convertEffectAccountTransferWithSchedule(
                            value
                        )

                    elif type(value) == AccountTransactionEffects.CredentialsUpdated:
                        _type.update({"contents": key})
                        result[key] = self.convertCredentialsUpdated(value)

                    elif type(value) == RegisteredData:
                        _type.update({"contents": key})
                        result[key] = self.convertType(value)

                    elif type(value) == AccountTransactionEffects.BakerConfigured:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectBakerConfigured(value)

                    elif type(value) == AccountTransactionEffects.DelegationConfigured:
                        _type.update({"contents": key})
                        result[key] = self.convertEffectDelegationConfigured(value)

                    elif type(value) in self.simple_types:
                        _type.update({"contents": key})
                        result[key] = self.convertType(value)

        return CCD_AccountTransactionEffects(**result), _type, _outcome

    def convertAccountTransactionDetails(
        self, message
    ) -> CCD_AccountTransactionDetails:
        result = {}
        for field, value in message.ListFields():
            key = field.name
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            if type(value) == AccountTransactionEffects:
                (
                    result[key],
                    _type,
                    result["outcome"],
                ) = self.convertAccountTransactionEffects(value)

        return CCD_AccountTransactionDetails(**result), CCD_TransactionType(**_type)

    def convertAccountCreationDetails(self, message) -> CCD_AccountCreationDetails:
        result = {}
        _type = {"type": "account_creation"}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if key == "credential_type":
                result[key] = value
                _type.update({"contents": CCD_CredentialType(value).name})

            elif type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_AccountCreationDetails(**result), CCD_TransactionType(**_type)

    def convertUpdatePayload(self, message) -> CCD_UpdatePayload:
        result = {}
        _type = {"type": "update"}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if self.valueIsEmpty(value):
                pass
            else:
                _type.update({"contents": key})

                if type(value) == ExchangeRate:
                    value_as_dict = MessageToDict(value)
                    result[key] = CCD_ExchangeRate(
                        **{
                            "numerator": value_as_dict["value"]["numerator"],
                            "denominator": value_as_dict["value"]["denominator"],
                        }
                    )

                elif type(value) in [BakerStakeThreshold, ProtocolUpdate]:
                    result[key] = self.convertTypeWithSingleValues(value)

                elif type(value) == Level1Update:
                    result[key] = self.convertLevel1Update(value)

                elif type(value) == IpInfo:
                    result[key] = self.convertIpInfo(value)

                # TODO: no test available
                elif type(value) == ElectionDifficulty:
                    result[key] = self.convertElectionDifficulty(value)

                # TODO: no test available
                elif type(value) == MintDistributionCpv0:
                    result[key] = self.convertMintDistributionCpv0(value)

                # TODO: no test available
                elif type(value) == TransactionFeeDistribution:
                    result[key] = self.convertTransactionFeeDistribution(value)

                # TODO: no test available
                elif type(value) == GasRewards:
                    result[key] = self.convertGasRewards(value)

                # TODO: no test available
                elif type(value) == RootUpdate:
                    result[key] = self.convertRootUpdate(value)

                # TODO: no test available
                elif type(value) == ArInfo:
                    result[key] = self.convertArInfo(value)

                # TODO: no test available
                elif type(value) == CooldownParametersCpv1:
                    result[key] = self.convertCooldownParametersCpv1(value)

                # TODO: no test available
                elif type(value) == PoolParametersCpv1:
                    result[key] = self.convertPoolParametersCpv1(value)

                # TODO: no test available
                elif type(value) == TimeParametersCpv1:
                    result[key] = self.convertTimeParametersCpv1(value)

                # TODO: no test available
                elif type(value) == MintDistributionCpv1:
                    result[key] = self.convertMintDistributionCpv1(value)

                elif type(value) == FinalizationCommitteeParameters:
                    result[key] = self.convertFinalizationCommitteeParameters(value)

                elif type(value) in self.simple_types:
                    result[key] = self.convertType(value)

        return CCD_UpdatePayload(**result), _type

    def convertUpdateDetails(self, message) -> CCD_UpdateDetails:
        result = {}
        _type = {"type": "update"}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if self.valueIsEmpty(value):
                pass
            else:
                if type(value) == UpdatePayload:
                    result[key], _type = self.convertUpdatePayload(value)

                elif type(value) in self.simple_types:
                    result[key] = self.convertType(value)

        return CCD_UpdateDetails(**result), CCD_TransactionType(**_type)

    def get_block_transaction_events(
        self: GRPCClient,
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> CCD_Block:
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        grpc_return_value = self.stub_on_net(
            net, "GetBlockTransactionEvents", blockHashInput
        )

        tx_list = []
        for tx in list(grpc_return_value):
            result = {}
            for field, value in tx.ListFields():
                key = field.name
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)

                if type(value) == UpdateDetails:
                    result[key], result["type"] = self.convertUpdateDetails(value)

                if type(value) == AccountCreationDetails:
                    result[key], result["type"] = self.convertAccountCreationDetails(
                        value
                    )

                if type(value) == AccountTransactionDetails:
                    result[key], result["type"] = self.convertAccountTransactionDetails(
                        value
                    )

            tx_list.append(CCD_BlockItemSummary(**result))

        return CCD_Block(**{"transaction_summaries": tx_list})



import io
import base58

from eth_hash.auto import keccak

from enum import Enum
from ccdexplorer_fundamentals.enums import NET

from rich.console import Console
from pymongo import ReplaceOne
from ccdexplorer_fundamentals.GRPCClient import GRPCClient

# from ccdexplorer_fundamentals.mongodb import Collections

console = Console()


class CNSActions(Enum):
    bid = "bid"
    cancel = "cancel"
    finalize = "finalize"
    setAddress = "setAddress"
    setData = "setData"
    transfer = "transfer"
    register = "register"
    createSubdomain = "createSubdomain"
    getTokenExpiry = "getTokenExpiry"
    extend = "extend"


class CNSEvent(Enum):
    CancelEvent = 233
    AbortEvent = 242
    FinalizeEvent = 243
    BidEvent = 244
    AuctionEvent = 247
    BuyEvent = 248
    UnlistEvent = 249


class CNSDomain:
    def __init__(self):
        self.function_calls = {}  # event["receiveName"] : event["message"]
        self.cns_event: CNSEvent = None
        self.transfer_to = None
        self.amount = 0
        self.subdomain = None
        self.tokenId = None
        self.action: CNSActions = None
        self.action_message = None
        self.domain_name = None
        self.set_address = None
        self.set_data_key = None
        self.set_data_value = None
        self.duration_years = None
        self.register_address = None

    # Helper functions
    def __repl__(self):
        return f"{self.tokenId=} | {self.domain_name=}"

    def bytes_from_hex_tokenID(self, hex):
        the_list = list(bytes.fromhex(hex[2:]))
        the_list.insert(0, 32)
        return the_list

    def write_binary_to_file(self, hex):
        newFileBytes = self.bytes_from_hex_tokenID(hex)
        newFileByteArray = bytearray(newFileBytes)
        newFile = open("token.bin", "wb")
        newFile.write(newFileByteArray)

    def namehash_dome(self, name: str, encoding="utf-8"):
        """ENS "namehash()" convention mapping of strings to bytes(32) hashes.

        Recursive function variant. Performs slightly better than the
        generator-based variant, but can't handle names with infinite (or
        extremely large) number of labels.

        :param name: name to hash, labels separated by dots
        :type name: str
        :returns: bytes(32)"""

        name = (
            name.encode("latin_1")
            .decode("raw_unicode_escape")
            .encode("utf-16", "surrogatepass")
            .decode("utf-16")
            .encode("raw_unicode_escape")
            .decode("latin_1")
        )

        if name == "":
            return b"\x00" * 32
        else:
            label, _, remainder = name.partition(".")
            return keccak(
                self.namehash_dome(remainder) + keccak(bytes(label, encoding=encoding))
            )

    def namehash_hashlib(self, name: str, encoding="utf-8"):
        """ENS "namehash()" convention mapping of strings to bytes(32) hashes.

        Recursive function variant. Performs slightly better than the
        generator-based variant, but can't handle names with infinite (or
        extremely large) number of labels.

        :param name: name to hash, labels separated by dots
        :type name: str
        :returns: bytes(32)"""
        import hashlib

        if name == "":
            return b"\x00" * 32
        else:
            label, _, remainder = name.partition(".")
            second_part = hashlib.new("sha3_256")
            second_part.update(bytes(label, encoding=encoding))
            ss = second_part.digest()
            return_statement = hashlib.new("sha3_256")
            return_statement.update(self.namehash_hashlib(remainder) + ss)
            return return_statement.digest()

    # CNS utilities
    def read_domain_owner(self, bs):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return (0, "Token doesn't exist.")
        elif t == 1:
            to_ = self.address(bs)
            return (1, to_)

    # CNS Defined terms
    def token(self, bs):
        contract_index, contract_subindex = self.contract_address(bs)
        tokenID = self.token_id(bs)
        return contract_index, contract_subindex, tokenID

    def account_address(self, bs):
        addr = bs.read(32)
        return base58.b58encode_check(b"\x01" + addr).decode()

    def contract_address(self, bs):
        return int.from_bytes(bs.read(8), byteorder="little"), int.from_bytes(
            bs.read(8), byteorder="little"
        )

    def royalty_length(self, bs):
        return int.from_bytes(bs.read(4), byteorder="little")

    def royalty(self, bs):
        beneficiary_ = self.account_address(bs)  # noqa: F841
        percentage = self.percentage(bs)  # noqa: F841

    def address(self, bs):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return self.account_address(bs)
        elif t == 1:
            return self.contract_address(bs)
        else:
            return "x" * 50

    def translate_hex_to_bytes(self, parameter_hex: str) -> bytes:
        parameter_bytes = list(bytes.fromhex(parameter_hex))
        parameter_bytes.insert(0, 32)
        return bytes(parameter_bytes)

    def get_cns_domain_name_v2(self, grpcclient: GRPCClient, hex: str, net: str):
        try:
            block_hash = "last_final"
            instance_index = 7073
            instance_subindex = 0
            entrypoint = "BictoryCnsNft.getTokenInfo"
            parameter_bytes = self.translate_hex_to_bytes(hex)

            ii = grpcclient.invoke_instance(
                block_hash,
                instance_index,
                instance_subindex,
                entrypoint,
                parameter_bytes,
                NET(net),
            )
            if ii.success is not None:
                res = ii.success.return_value
                bb = list(bytes.fromhex(res.decode()))
                domain_name = bytes(bb[5:-8]).decode()
        except Exception as _:
            domain_name = "Unable to determine..."
        self.domain_name = domain_name

    def get_domain_name_owner_v2(
        self, grpcclient: GRPCClient, mongodb, Collections, domain_name
    ):
        console.log(f"get_domain_name_owner for {domain_name}")
        result = mongodb.mainnet[Collections.cns_domains].find_one(
            {"domain_name": domain_name}
        )
        hex = None
        if result:
            hex = result["_id"]

        if not hex:
            bb = self.namehash_dome(domain_name)
            hex = bytes.hex(bb)
            document_to_store = ReplaceOne(
                {
                    "_id": hex,
                },
                {
                    "_id": hex,
                    "domain_name": domain_name,
                },
                upsert=True,
            )
            result = mongodb.mainnet[Collections.cns_domains].bulk_write(
                [document_to_store]
            )

        block_hash = "last_final"
        instance_index = 7073
        instance_subindex = 0
        entrypoint = "BictoryCnsNft.getTokenExpiry"
        parameter_bytes = self.translate_hex_to_bytes(hex)

        ii = grpcclient.invoke_instance(
            block_hash,
            instance_index,
            instance_subindex,
            entrypoint,
            parameter_bytes,
        )

        if ii.success is not None:
            res = ii.success.return_value
            bb = list(bytes.fromhex(res.decode()))
            byt = io.BytesIO(bytes(bb))
            (status, to) = CNSDomain().read_domain_owner(byt)
            return (status, to)
        else:
            return (-1, None)

    def string(self, bs):
        # bs = io.BytesIO(bytes.fromhex(hex))
        n = int.from_bytes(bs.read(4), byteorder="little")
        name = bs.read(n)
        return bytes.decode(name, "UTF-8")

    def data_value(self, bs):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return ""
        elif t == 1:
            return self.address(bs)
        elif t == 2:
            return self.string(bs)
        elif t == 3:
            return bs
        elif t == 4:
            return self.string(bs)
        elif t == 5:
            return self.token()
        else:
            raise Exception("invalid type")

    def decode_set_data_from(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        domain_ = self.string(bs)
        key_ = self.string(bs)
        data_value_ = self.data_value(bs)
        return domain_, key_, data_value_

    def decode_set_address_from(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        domain_ = self.string(bs)
        address_ = self.address(bs)
        return domain_, address_

    def decode_from_extend(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        name_ = self.string(bs)
        duration_ = int.from_bytes(bs.read(1), byteorder="little")
        return name_, duration_

    def decode_from_register(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        name_ = self.string(bs)
        address_ = self.address(bs)
        duration_ = int.from_bytes(bs.read(1), byteorder="little")
        return name_, address_, duration_

    def decode_domain_from(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        name_ = self.string(bs)
        return name_

    def decode_token_id_from(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        id_ = self.token_id(bs)
        return id_

    def decode_subdomain_from(self, hex):
        bs = io.BytesIO(bytes.fromhex(hex))
        n = int.from_bytes(bs.read(4), byteorder="little")
        name = bs.read(n)
        return bytes.decode(name, "UTF-8")

    def decode_transfer_to_from(self, hex):
        all_data = self.transfer_parameter(io.BytesIO(bytes.fromhex(hex)))
        if len(all_data) > 0:
            if len(all_data[0]) > 3:
                return all_data[0][0], all_data[0][3]  # tokenId, to_address
            else:
                return None
        else:
            return None

    def transfer_parameter(self, bs):
        n = int.from_bytes(bs.read(2), byteorder="little")
        return self.transfers(bs, n)

    def transfers(self, bs, n):
        return [self.transfer(bs) for _ in range(n)]

    def transfer(self, bs):
        id_ = self.token_id(bs)
        amount = self.token_amount(bs)
        from_ = self.address(bs)
        to = self.receiver(bs)
        data = self.additional_data(bs)
        return [id_, amount, from_, to, data]

    def token_id(self, bs):
        n = int.from_bytes(bs.read(1), byteorder="little")
        return bytes.hex(bs.read(n))

    def token_amount(self, bs):
        return int.from_bytes(bs.read(1), byteorder="little")

    def percentage(self, bs):
        return int.from_bytes(bs.read(8), byteorder="little")

    def receiver(self, bs):
        t = int.from_bytes(bs.read(1), byteorder="little")
        if t == 0:
            return self.account_address(bs)
        elif t == 1:
            return self.contract_address(bs), self.receive_hook_name(bs)
        else:
            return "x" * 50

    def receive_hook_name(self, bs):
        n = int.from_bytes(bs.read(2), byteorder="little")
        name = bs.read(n)
        return bytes.decode(name, "UTF-8")

    def additional_data(self, bs):
        n = int.from_bytes(bs.read(2), byteorder="little")
        data = bs.read(n)
        return data

    # CNS Event
    def buyEvent(self, hexParameter):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        self.cns_event = CNSEvent(tag_)

        contract_index, contract_subindex, token_id_ = self.token(bs)
        seller_ = self.account_address(bs)
        buyer_ = self.account_address(bs)
        price_ = self.token_amount(bs)
        seller_share_ = self.token_amount(bs)
        royalty_len_ = int.from_bytes(bs.read(4), byteorder="little")
        royalties_ = []
        for _ in range(royalty_len_):
            royalties_.append(self.royalty(bs))

        return [
            tag_,
            contract_index,
            contract_subindex,
            token_id_,
            seller_,
            buyer_,
            price_,
            seller_share_,
            royalties_,
        ]

    def bidEvent(self, hexParameter):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        self.cns_event = CNSEvent(tag_)

        contract_index, contract_subindex, token_id_ = self.token(bs)
        bidder_ = self.account_address(bs)
        amount_ = self.token_amount(bs)

        return [tag_, contract_index, contract_subindex, token_id_, bidder_, amount_]

    def cancelEvent(self, hexParameter):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        self.cns_event = CNSEvent(tag_)

        contract_index, contract_subindex, token_id_ = self.token(bs)
        owner_ = self.account_address(bs)

        return [tag_, contract_index, contract_subindex, token_id_, owner_]

    def abortEvent(self, hexParameter):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        self.cns_event = CNSEvent(tag_)

        contract_index, contract_subindex, token_id_ = self.token(bs)
        owner_ = self.account_address(bs)
        bidder_ = self.account_address(bs)
        amount_ = self.token_amount(bs)

        return [
            tag_,
            contract_index,
            contract_subindex,
            token_id_,
            owner_,
            bidder_,
            amount_,
        ]

    def finalizeEvent(self, hexParameter):
        bs = io.BytesIO(bytes.fromhex(hexParameter))

        tag_ = int.from_bytes(bs.read(1), byteorder="little")
        self.cns_event = CNSEvent(tag_)

        contract_index, contract_subindex, token_id_ = self.token(bs)
        seller_ = self.account_address(bs)
        winner_ = self.account_address(bs)
        price_ = self.token_amount(bs)
        seller_share = self.token_amount(bs)
        royalty_length_ = self.royalty_length(bs)
        royalties = [self.royalty(bs) for x in range(royalty_length_)]

        return [
            tag_,
            contract_index,
            contract_subindex,
            token_id_,
            seller_,
            winner_,
            price_,
            seller_share,
            royalty_length_,
            royalties,
        ]

    def finalize(self, hexParameter):
        bs = io.BytesIO(bytes.fromhex(hexParameter))
        tag_ = int.from_bytes(bs.read(1), byteorder="little")

        # only available for finalizeEvent
        seller_, winner_, price_, seller_share, royalty_length_, royalties = (
            None,
            None,
            None,
            None,
            None,
            None,
        )

        # only available for abortEvent, cancelEvent
        owner_, bidder_, amount_ = None, None, None

        # tag 161 is undocumented for now?
        (
            contract_index,
            contract_subindex,
            token_id_,
        ) = (
            None,
            None,
            None,
        )

        if tag_ == 243:
            (
                tag_,
                contract_index,
                contract_subindex,
                token_id_,
                seller_,
                winner_,
                price_,
                seller_share,
                royalty_length_,
                royalties,
            ) = self.finalizeEvent(hexParameter)

        if tag_ == 242:
            (
                tag_,
                contract_index,
                contract_subindex,
                token_id_,
                owner_,
                bidder_,
                amount_,
            ) = self.abortEvent(hexParameter)

        if tag_ == 233:
            (
                tag_,
                contract_index,
                contract_subindex,
                token_id_,
                owner_,
            ) = self.cancelEvent(hexParameter)

        return [
            tag_,
            contract_index,
            contract_subindex,
            token_id_,
            seller_,
            winner_,
            price_,
            seller_share,
            royalty_length_,
            royalties,
            owner_,
            bidder_,
            amount_,
        ]
